from collections.abc import Awaitable, Callable
from functools import partial
from typing import Final, final, override

from wirio._service_lookup._call_site_kind import CallSiteKind
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._service_call_site import ServiceCallSite
from wirio._service_lookup._typed_type import TypedType
from wirio.wirio_undefined import WirioUndefined


@final
class AsyncFactoryCallSite(ServiceCallSite):
    _service_type: Final[TypedType]
    _implementation_factory: Callable[..., Awaitable[object]]

    def __init__(
        self,
        cache: ResultCache,
        service_type: TypedType,
        service_key: object | None = WirioUndefined.INSTANCE,
    ) -> None:
        is_keyed_implementation_factory = service_key is not WirioUndefined.INSTANCE
        service_key_to_add = service_key if is_keyed_implementation_factory else None
        super().__init__(cache=cache, key=service_key_to_add)
        self._service_type = service_type

    @classmethod
    def from_implementation_factory(
        cls,
        cache: ResultCache,
        service_type: TypedType,
        implementation_factory: Callable[..., Awaitable[object]],
    ) -> "AsyncFactoryCallSite":
        self = cls(cache=cache, service_type=service_type)
        self._implementation_factory = implementation_factory
        return self

    @classmethod
    def from_keyed_implementation_factory(
        cls,
        cache: ResultCache,
        service_type: TypedType,
        implementation_factory: Callable[..., Awaitable[object]],
        service_key: object | None,
    ) -> "AsyncFactoryCallSite":
        self = cls(cache=cache, service_type=service_type, service_key=service_key)
        self._implementation_factory = partial(implementation_factory, service_key)
        return self

    @property
    @override
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    @override
    def kind(self) -> CallSiteKind:
        return CallSiteKind.ASYNC_FACTORY

    @property
    def implementation_factory(
        self,
    ) -> Callable[..., Awaitable[object]]:
        return self._implementation_factory



from typing import Final, final, override

from wirio._service_lookup._call_site_kind import CallSiteKind
from wirio._service_lookup._constructor_information import (
    ConstructorInformation,
)
from wirio._service_lookup._parameter_information import (
    ParameterInformation,
)
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._service_call_site import ServiceCallSite
from wirio._service_lookup._typed_type import TypedType


@final
class ConstructorCallSite(ServiceCallSite):
    _service_type: Final[TypedType]
    _constructor_information: Final[ConstructorInformation]
    _parameters: Final[list[ParameterInformation]]
    _parameter_call_sites: Final[list[ServiceCallSite | None]]

    def __init__(  # noqa: PLR0913
        self,
        cache: ResultCache,
        service_type: TypedType,
        constructor_information: ConstructorInformation,
        parameters: list[ParameterInformation],
        parameter_call_sites: list[ServiceCallSite | None],
        service_key: object | None = None,
    ) -> None:
        super().__init__(cache=cache, key=service_key)
        self._service_type = service_type
        self._constructor_information = constructor_information
        self._parameters = parameters
        self._parameter_call_sites = parameter_call_sites

    @property
    @override
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    @override
    def kind(self) -> CallSiteKind:
        return CallSiteKind.CONSTRUCTOR

    @property
    def constructor_information(self) -> ConstructorInformation:
        return self._constructor_information

    @property
    def parameters(self) -> list[ParameterInformation]:
        return self._parameters

    @property
    def parameter_call_sites(self) -> list[ServiceCallSite | None]:
        return self._parameter_call_sites



from abc import ABC, abstractmethod

from wirio._service_lookup._typed_type import (
    TypedType,
)


class KeyedServiceProvider(ABC):
    """Retrieve services using a key and a type."""

    @abstractmethod
    async def get_keyed_service_object(
        self, service_key: object | None, service_type: TypedType
    ) -> object | None: ...

    @abstractmethod
    async def get_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        ...

    @abstractmethod
    async def get_required_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise an error."""
        ...



import typing
from collections.abc import (
    Hashable,
)
from typing import Any, Final, final, override


@final
class TypedType(Hashable):
    """Version of :class:`type` that takes into account generic parameters."""

    _origin: Final[Any]
    _args: Final[tuple[Any, ...]]

    def __init__(
        self,
        type_: Any,  # noqa: ANN401
    ) -> None:
        origin = typing.get_origin(type_)
        has_no_generics = origin is None

        if has_no_generics:
            self._origin = type_
            self._args = typing.get_args(type_)
            return

        self._origin = origin
        self._args = typing.get_args(type_)

    @classmethod
    def from_type(cls, type_: type) -> "TypedType":
        return cls(type_)

    @classmethod
    def from_instance(cls, instance: object) -> "TypedType":
        instance_type = getattr(instance, "__orig_class__", None)

        if instance_type is None:
            error_message = "The instance does not retain type hint information because it has no generics"
            raise ValueError(error_message)

        return cls(instance_type)

    def to_type(self) -> type:
        return self._origin

    def invoke(self, parameter_values: list[object]) -> object:
        has_parameters = len(parameter_values) > 0
        has_generics = len(self._args) > 0

        if not has_parameters:
            if not has_generics:
                return self._origin()

            return self._origin[*self._args]()  # pyright: ignore[reportIndexIssue, reportUnknownVariableType]

        if not has_generics:
            return self._origin(*parameter_values)

        return self._origin[*self._args](*parameter_values)  # pyright: ignore[reportIndexIssue, reportUnknownVariableType]

    def _create_representation(
        self,
        origin: Any,  # noqa: ANN401
        args: tuple[Any, ...],
    ) -> str:
        args_representation = ""

        if len(args) > 0:
            for arg in args:
                arg_origin = typing.get_origin(arg)
                arg_args = typing.get_args(arg)
                has_generics = arg_origin is not None

                if has_generics:
                    args_representation += self._create_representation(
                        arg_origin, arg_args
                    )
                else:
                    args_representation += f"{arg.__module__}.{arg.__qualname__}"

                if arg != args[-1]:
                    args_representation += ", "

        if len(args_representation) > 0:
            args_representation = f"[{args_representation}]"

        return f"{origin.__module__}.{origin.__qualname__}{args_representation}"

    @override
    def __repr__(self) -> str:
        return self._create_representation(self._origin, self._args)

    @override
    def __hash__(self) -> int:
        return hash(self._origin) ^ hash(self._args)

    @override
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, TypedType):
            return NotImplemented

        return self._origin == value._origin and self._args == value._args



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable

from wirio._service_lookup._service_call_site import (
    ServiceCallSite,
)
from wirio.service_provider_engine_scope import (
    ServiceProviderEngineScope,
)


class ServiceProviderEngine(ABC):
    @abstractmethod
    def realize_service(
        self, call_site: ServiceCallSite
    ) -> Callable[[ServiceProviderEngineScope], Awaitable[object | None]]: ...



from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager

from wirio.abstractions.base_service_provider import (
    BaseServiceProvider,
)


class ServiceScope(AbstractAsyncContextManager["ServiceScope"], ABC):
    """Defines a disposable service scope.

    The __aexit__ method ends the scope lifetime. Once called, any scoped
    services that have been resolved from ServiceProvider will be disposed.
    """

    @property
    @abstractmethod
    def service_provider(self) -> BaseServiceProvider:
        """Gets the :class:`BaseServiceProvider` used to resolve dependencies from the scope."""

    @abstractmethod
    async def get_service[TService](
        self, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        ...

    @abstractmethod
    async def get_required_service[TService](
        self, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise :class:`NoServiceRegisteredError`."""
        ...

    @abstractmethod
    async def get_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        ...

    @abstractmethod
    async def get_required_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise an error."""
        ...



from abc import ABC, abstractmethod
from typing import cast

from wirio._service_lookup._typed_type import TypedType
from wirio.abstractions.keyed_service import KeyedService
from wirio.abstractions.keyed_service_provider import (
    KeyedServiceProvider,
)
from wirio.abstractions.service_scope_factory import (
    ServiceScopeFactory,
)
from wirio.exceptions import (
    KeyedServiceAnyKeyUsedToResolveServiceError,
    NoKeyedServiceRegisteredError,
    NoServiceRegisteredError,
)


class BaseServiceProvider(KeyedServiceProvider, ServiceScopeFactory, ABC):
    """Define a mechanism for retrieving a service object; that is, an object that provides custom support to other objects."""

    @abstractmethod
    async def get_service_object(self, service_type: TypedType) -> object | None: ...

    @abstractmethod
    async def get_keyed_service_object(
        self, service_key: object | None, service_type: TypedType
    ) -> object | None: ...

    async def get_service[TService](
        self, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        service = await self.get_service_object(TypedType.from_type(service_type))

        if service is None:
            return None

        return cast("TService", service)

    async def get_required_service[TService](
        self, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise :class:`NoServiceRegisteredError`."""
        service = await self.get_service(service_type)

        if service is None:
            raise NoServiceRegisteredError(TypedType.from_type(service_type))

        return service

    async def get_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        if service_key is KeyedService.ANY_KEY:
            raise KeyedServiceAnyKeyUsedToResolveServiceError

        service = await self.get_keyed_service_object(
            service_key, TypedType.from_type(service_type)
        )

        if service is None:
            return None

        return cast("TService", service)

    async def get_required_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise an error."""
        if service_key is KeyedService.ANY_KEY:
            raise KeyedServiceAnyKeyUsedToResolveServiceError

        service = await self.get_keyed_service(service_key, service_type)

        if service is None:
            if service_key is None:
                raise NoServiceRegisteredError(TypedType.from_type(service_type))

            raise NoKeyedServiceRegisteredError(
                service_type=TypedType.from_type(service_type),
                service_key_type=type(service_key),
            )

        return service



from types import TracebackType
from typing import TYPE_CHECKING, Final, Self, final, override

from wirio._service_lookup._asyncio_reentrant_lock import (
    AsyncioReentrantLock,
)
from wirio._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from wirio._service_lookup._supports_async_context_manager import (
    SupportsAsyncContextManager,
)
from wirio._service_lookup._supports_context_manager import (
    SupportsContextManager,
)
from wirio._service_lookup._typed_type import TypedType
from wirio._service_lookup.service_cache_key import ServiceCacheKey
from wirio.abstractions.base_service_provider import BaseServiceProvider
from wirio.abstractions.service_scope import ServiceScope
from wirio.exceptions import ObjectDisposedError

if TYPE_CHECKING:
    from wirio.service_provider import ServiceProvider


@final
class ServiceProviderEngineScope(BaseServiceProvider, ServiceScope):
    """Container resolving services with scope."""

    _root_provider: Final["ServiceProvider"]
    _is_root_scope: Final[bool]
    _is_disposed: bool
    _disposables: list[object] | None
    _resolved_services: Final[dict[ServiceCacheKey, object | None]]

    # A reentrant lock is needed when the lifetime is scoped and the service has a context manager
    _resolved_services_lock: Final[AsyncioReentrantLock]

    def __init__(
        self, service_provider: "ServiceProvider", is_root_scope: bool
    ) -> None:
        self._root_provider = service_provider
        self._is_root_scope = is_root_scope
        self._is_disposed = False
        self._disposables = None
        self._resolved_services = {}
        self._resolved_services_lock = AsyncioReentrantLock()

    @property
    def root_provider(self) -> "ServiceProvider":
        return self._root_provider

    @property
    def is_root_scope(self) -> bool:
        return self._is_root_scope

    @property
    def resolved_services(self) -> dict[ServiceCacheKey, object | None]:
        return self._resolved_services

    @property
    def resolved_services_lock(self) -> AsyncioReentrantLock:
        """Protect the state on the scope.

        In particular, for the root scope, it protects the list of disposable entries only, since :attr:`resolved_services` are cached on :class:`CallSites`.

        For other scopes, it protects :attr:`resolved_services` and the list of disposables.
        """
        return self._resolved_services_lock

    @property
    @override
    def service_provider(self) -> BaseServiceProvider:
        return self

    @override
    def create_scope(self) -> ServiceScope:
        return self._root_provider.create_scope()

    @override
    async def get_service_object(self, service_type: TypedType) -> object | None:
        if self._is_disposed:
            raise ObjectDisposedError

        return await self._root_provider.get_service_from_service_identifier(
            service_identifier=ServiceIdentifier.from_service_type(service_type),
            service_provider_engine_scope=self,
        )

    @override
    async def get_keyed_service_object(
        self, service_key: object | None, service_type: TypedType
    ) -> object | None:
        if self._is_disposed:
            raise ObjectDisposedError

        return await self._root_provider.get_service_from_service_identifier(
            service_identifier=ServiceIdentifier.from_service_type(
                service_type=service_type, service_key=service_key
            ),
            service_provider_engine_scope=self,
        )

    async def get_service[TService](
        self, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        return await super().get_service(service_type)

    async def get_required_service[TService](
        self, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise :class:`NoServiceRegisteredError`."""
        return await super().get_required_service(service_type)

    async def get_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService | None:
        """Get service of type `TService` or return `None`."""
        return await super().get_keyed_service(
            service_key=service_key, service_type=service_type
        )

    async def get_required_keyed_service[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService:
        """Get service of type `TService` or raise an error."""
        return await super().get_required_keyed_service(
            service_key=service_key, service_type=service_type
        )

    async def capture_disposable(self, service: object | None) -> object | None:
        if service is self or not (
            isinstance(service, (SupportsAsyncContextManager, SupportsContextManager))
        ):
            return service

        is_disposed = False

        async with self._resolved_services_lock:
            if self._is_disposed:
                is_disposed = True
            else:
                if self._disposables is None:
                    self._disposables = []

                self._disposables.append(service)

        # Don't run customer code under the lock
        if is_disposed:
            if isinstance(service, SupportsAsyncContextManager):
                await service.__aexit__(None, None, None)
            else:
                service.__exit__(None, None, None)

            raise ObjectDisposedError

        return service

    async def _begin_dispose(self) -> list[object] | None:
        async with self._resolved_services_lock:
            if self._is_disposed:
                return None

            # We've transitioned to the disposed state, so future calls to
            # :meth:`capture_disposable` will immediately dispose the object.
            # No further changes to disposables are allowed.
            self._is_disposed = True

        if self._is_root_scope and not self._root_provider.is_disposed:
            # If this :class:`ServiceProviderEngineScope` instance is a root scope, disposing this instance will need to dispose the :attr:`root_provider` too.
            # Otherwise the :attr:`root_provider` will never get disposed and will leak.
            # Note, if the :attr:`root_provider` gets disposed first, it will automatically dispose all attached :class:`ServiceProviderEngineScope` objects.
            await self._root_provider.__aexit__(None, None, None)

        # :attr:`_resolved_services` is never cleared for singletons because there might be a compilation running in background
        # trying to get a cached singleton service. If it doesn't find it
        # it will try to create a new one which will result in an :class:`ObjectDisposedError`.
        return self._disposables

    @override
    async def __aenter__(self) -> Self:
        if not self._root_provider.is_fully_initialized:
            await self._root_provider.__aenter__()

        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        to_dispose = await self._begin_dispose()

        if to_dispose is None:
            return None

        for i in range(len(to_dispose) - 1, -1, -1):
            service = to_dispose[i]

            if isinstance(service, SupportsAsyncContextManager):
                await service.__aexit__(None, None, None)
            elif isinstance(service, SupportsContextManager):
                service.__exit__(None, None, None)



from fastapi import FastAPI

from wirio.service_container import ServiceContainer
from wirio.service_provider import ServiceProvider


def get_service_provider(app: FastAPI) -> ServiceProvider:
    assert isinstance(app.state.wirio_service_provider, ServiceProvider)
    return app.state.wirio_service_provider


def get_service_container(app: FastAPI) -> ServiceContainer:
    assert isinstance(app.state.wirio_services, ServiceContainer)
    return app.state.wirio_services



from collections.abc import Awaitable, Callable
from functools import partial
from typing import Final, Self, override

from wirio._service_lookup._typed_type import TypedType
from wirio.exceptions import NonKeyedDescriptorMisuseError
from wirio.service_lifetime import ServiceLifetime


class ServiceDescriptor:
    """Service registration."""

    _service_type: Final[TypedType]
    _lifetime: Final[ServiceLifetime]
    _implementation_type: TypedType | None
    _implementation_instance: object | None
    _sync_implementation_factory: Callable[..., object] | None
    _async_implementation_factory: Callable[..., Awaitable[object]] | None
    _service_key: object | None
    _auto_activate: bool

    def __init__(
        self,
        service_type: type,
        service_key: object | None,
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> None:
        self._service_type = TypedType.from_type(service_type)
        self._service_key = service_key
        self._lifetime = lifetime
        self._auto_activate = auto_activate
        self._implementation_type = None
        self._implementation_instance = None
        self._sync_implementation_factory = None
        self._async_implementation_factory = None

    @classmethod
    def from_implementation_type(
        cls,
        service_type: type,
        implementation_type: type,
        service_key: object | None,
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=service_key,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )
        self._implementation_type = TypedType.from_type(implementation_type)
        return self

    @classmethod
    def from_implementation_instance(
        cls,
        service_type: type,
        implementation_instance: object,
        service_key: object | None,
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=service_key,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )
        self._implementation_instance = implementation_instance
        return self

    @classmethod
    def from_sync_implementation_factory(
        cls,
        service_type: type,
        implementation_factory: Callable[..., object],
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=None,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )
        self._sync_implementation_factory = implementation_factory
        return self

    @classmethod
    def from_keyed_sync_implementation_factory(
        cls,
        service_type: type,
        implementation_factory: Callable[..., object],
        service_key: object | None,
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=service_key,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )

        if service_key is None:
            none_keyed_implementation_factory = partial(implementation_factory, None)
            self._sync_implementation_factory = none_keyed_implementation_factory
        else:
            self._sync_implementation_factory = implementation_factory

        return self

    @classmethod
    def from_async_implementation_factory(
        cls,
        service_type: type,
        implementation_factory: Callable[..., Awaitable[object]],
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=None,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )
        self._async_implementation_factory = implementation_factory
        return self

    @classmethod
    def from_keyed_async_implementation_factory(
        cls,
        service_type: type,
        implementation_factory: Callable[..., Awaitable[object]],
        service_key: object | None,
        lifetime: ServiceLifetime,
        auto_activate: bool,
    ) -> Self:
        self = cls(
            service_type=service_type,
            service_key=service_key,
            lifetime=lifetime,
            auto_activate=auto_activate,
        )

        if service_key is None:
            none_keyed_implementation_factory = partial(implementation_factory, None)
            self._async_implementation_factory = none_keyed_implementation_factory
        else:
            self._async_implementation_factory = implementation_factory

        return self

    @property
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    def service_key(self) -> object | None:
        return self._service_key

    @property
    def lifetime(self) -> ServiceLifetime:
        return self._lifetime

    @property
    def auto_activate(self) -> bool:
        return self._auto_activate

    @auto_activate.setter
    def auto_activate(self, value: bool) -> None:
        self._auto_activate = value

    @property
    def implementation_type(self) -> TypedType | None:
        return self._implementation_type

    @property
    def implementation_instance(self) -> object | None:
        return self._implementation_instance

    @property
    def sync_implementation_factory(
        self,
    ) -> Callable[..., object] | None:
        if self.is_keyed_service:
            return None

        return self._sync_implementation_factory

    @property
    def async_implementation_factory(
        self,
    ) -> Callable[..., Awaitable[object]] | None:
        if self.is_keyed_service:
            return None

        return self._async_implementation_factory

    @property
    def keyed_sync_implementation_factory(
        self,
    ) -> Callable[..., object] | None:
        """Get the factory used for creating keyed synchronous service instances, or raise :class:`NonKeyedDescriptorMisuseError` if :attr:`is_keyed_service` is `False`."""
        if not self.is_keyed_service:
            raise NonKeyedDescriptorMisuseError

        return self._sync_implementation_factory

    @property
    def keyed_async_implementation_factory(
        self,
    ) -> Callable[..., Awaitable[object]] | None:
        """Get the factory used for creating keyed asynchronous service instances, or raise :class:`NonKeyedDescriptorMisuseError` if :attr:`is_keyed_service` is `False`."""
        if not self.is_keyed_service:
            raise NonKeyedDescriptorMisuseError

        return self._async_implementation_factory

    @property
    def is_keyed_service(self) -> bool:
        return self._service_key is not None

    @property
    def keyed_implementation_type(self) -> TypedType | None:
        """Get the type that implements the service, or raise :class:`NonKeyedDescriptorMisuseError` if :attr:`is_keyed_service` is `False`."""
        if not self.is_keyed_service:
            raise NonKeyedDescriptorMisuseError

        return self._implementation_type

    @property
    def keyed_implementation_instance(self) -> object | None:
        """Get the instance that implements the service, or raise :class:`NonKeyedDescriptorMisuseError` if :attr:`is_keyed_service` is `False`."""
        if not self.is_keyed_service:
            raise NonKeyedDescriptorMisuseError

        return self._implementation_instance

    def has_implementation_type(self) -> bool:
        return self.get_implementation_type() is not None

    def get_implementation_type(self) -> TypedType | None:
        if self.is_keyed_service:
            return self.keyed_implementation_type

        return self._implementation_type

    def has_implementation_instance(self) -> bool:
        return self.get_implementation_instance() is not None

    def get_implementation_instance(self) -> object | None:
        if self.is_keyed_service:
            return self.keyed_implementation_instance

        return self._implementation_instance

    @override
    def __str__(self) -> str:  # noqa: PLR0911
        lifetime = f"service_type: {self.service_type}, lifetime: {self.lifetime}, "

        if self.is_keyed_service:
            lifetime += f"service_key: {self.service_key}, "

            if self.keyed_implementation_type is not None:
                return (
                    lifetime
                    + f"keyed_implementation_type: {self.keyed_implementation_type}"
                )

            if self.keyed_async_implementation_factory is not None:
                return (
                    lifetime
                    + f"keyed_async_implementation_factory: {self.keyed_async_implementation_factory}"
                )

            if self.keyed_sync_implementation_factory is not None:
                return (
                    lifetime
                    + f"keyed_sync_implementation_factory: {self.keyed_sync_implementation_factory}"
                )

            return (
                lifetime
                + f"keyed_implementation_instance: {self.keyed_implementation_instance}"
            )

        if self.implementation_type is not None:
            return lifetime + f"implementation_type: {self.implementation_type}"

        if self.async_implementation_factory is not None:
            return (
                lifetime
                + f"async_implementation_factory: {self.async_implementation_factory}"
            )

        if self.sync_implementation_factory is not None:
            return (
                lifetime
                + f"sync_implementation_factory: {self.sync_implementation_factory}"
            )

        return lifetime + (f"implementation_instance: {self.implementation_instance}")



# MIT License

# Copyright (c) 2023 Joshua George Albert
# Copyright (c) 2026 Andreu Codina

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# https://github.com/Joshuaalbert/FairAsyncRLock

import asyncio
from asyncio import CancelledError, Event, Task
from collections import deque
from contextlib import AbstractAsyncContextManager
from types import TracebackType
from typing import Any, Self, override


class AsyncioReentrantLock(AbstractAsyncContextManager["AsyncioReentrantLock"]):
    _owner: Task[Any] | None
    _count: int
    _owner_transfer: bool
    _queue: deque[Event]

    def __init__(self) -> None:
        self._owner = None
        self._count = 0
        self._owner_transfer = False
        self._queue = deque()

    @property
    def owner(self) -> Task[Any] | None:
        return self._owner

    @property
    def count(self) -> int:
        return self._count

    @property
    def queue(self) -> deque[Event]:
        return self._queue

    @property
    def is_locked(self) -> bool:
        return self._owner is not None

    def is_owner(self, task: Task[Any] | None = None) -> bool:
        if task is None:
            task = asyncio.current_task()

        return self._owner == task

    async def acquire(self) -> None:
        current_task = asyncio.current_task()

        # If the lock is reentrant, acquire it immediately
        if self.is_owner(task=current_task):
            self._count += 1
            return

        # If the lock is free (and ownership not in midst of transfer), acquire it immediately
        if self._count == 0 and not self._owner_transfer:
            self._owner = current_task
            self._count = 1
            return

        # Create an event for this task, to notify when it's ready for acquire
        event = Event()
        self._queue.append(event)

        # Wait for the lock to be free, then acquire
        try:
            await event.wait()
            self._owner_transfer = False
            self._owner = current_task
            self._count = 1
        except CancelledError:
            try:
                # If in queue, then cancelled before release
                self._queue.remove(event)
            except ValueError:
                # Otherwise, release happened, this was next, and we simulate passing on
                self._owner_transfer = False
                self._owner = current_task
                self._count = 1
                self._current_task_release()
            raise

    def _current_task_release(self) -> None:
        self._count -= 1
        if self._count == 0:
            self._owner = None

            if self._queue:
                # Wake up the next task in the queue
                event = self._queue.popleft()
                event.set()

                # Setting this here prevents another task getting lock until owner transfer
                self._owner_transfer = True

    def release(self) -> None:
        """Release the lock."""
        current_task = asyncio.current_task()

        if self._owner is None:
            error_message = (
                f"Cannot release un-acquired lock. {current_task} tried to release."
            )
            raise RuntimeError(error_message)

        if not self.is_owner(task=current_task):
            error_message = f"Cannot release foreign lock. {current_task} tried to unlock {self._owner}."
            raise RuntimeError(error_message)

        self._current_task_release()

    @override
    async def __aenter__(self) -> Self:
        await self.acquire()
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        self.release()



from typing import Final, final

from wirio._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from wirio._service_lookup._typed_type import TypedType
from wirio.exceptions import CircularDependencyError


@final
class _ChainItemInformation:
    _order: Final[int]
    _implementation_type: Final[TypedType | None]

    def __init__(self, order: int, implementation_type: TypedType | None) -> None:
        self._order = order
        self._implementation_type = implementation_type


@final
class CallSiteChain:
    """Keep track of the current resolution path to detect errors."""

    _call_site_chain: Final[dict[ServiceIdentifier, _ChainItemInformation]]

    def __init__(self) -> None:
        self._call_site_chain = {}

    def add(
        self,
        service_identifier: ServiceIdentifier,
        implementation_type: TypedType | None = None,
    ) -> None:
        self._call_site_chain[service_identifier] = _ChainItemInformation(
            order=len(self._call_site_chain), implementation_type=implementation_type
        )

    def remove(self, service_identifier: ServiceIdentifier) -> None:
        del self._call_site_chain[service_identifier]

    def check_circular_dependency(self, service_identifier: ServiceIdentifier) -> None:
        if service_identifier in self._call_site_chain:
            raise CircularDependencyError(service_identifier.service_type)



from abc import ABC, abstractmethod

from wirio._service_lookup._asyncio_reentrant_lock import AsyncioReentrantLock
from wirio._service_lookup._call_site_kind import CallSiteKind
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._typed_type import TypedType


class ServiceCallSite(ABC):
    """Representation of how a service must be created."""

    _cache: ResultCache
    _value: object | None
    _key: object | None
    _lock: AsyncioReentrantLock

    def __init__(
        self, cache: ResultCache, key: object | None, value: object | None = None
    ) -> None:
        self._cache = cache
        self._key = key
        self._value = value
        self._lock = AsyncioReentrantLock()

    @property
    def cache(self) -> ResultCache:
        return self._cache

    @property
    def key(self) -> object | None:
        return self._key

    @property
    def value(self) -> object | None:
        return self._value

    @value.setter
    def value(self, value: object | None) -> None:
        self._value = value

    @property
    def lock(self) -> AsyncioReentrantLock:
        return self._lock

    @property
    @abstractmethod
    def service_type(self) -> TypedType: ...

    @property
    @abstractmethod
    def kind(self) -> CallSiteKind: ...



from collections.abc import Hashable
from typing import (
    Final,
    final,
    override,
)

from wirio._service_lookup._typed_type import TypedType
from wirio.service_descriptor import ServiceDescriptor


@final
class ServiceIdentifier(Hashable):
    """Internal registered service during resolution."""

    _service_key: Final[object | None]
    _service_type: Final[TypedType]

    def __init__(self, service_key: object | None, service_type: TypedType) -> None:
        self._service_key = service_key
        self._service_type = service_type

    @property
    def service_key(self) -> object | None:
        return self._service_key

    @property
    def service_type(self) -> TypedType:
        return self._service_type

    @classmethod
    def from_service_type(
        cls, service_type: TypedType, service_key: object | None = None
    ) -> "ServiceIdentifier":
        return cls(service_key=service_key, service_type=service_type)

    @classmethod
    def from_descriptor(
        cls, service_descriptor: ServiceDescriptor
    ) -> "ServiceIdentifier":
        return cls(
            service_key=service_descriptor.service_key,
            service_type=service_descriptor.service_type,
        )

    @override
    def __hash__(self) -> int:
        if self.service_key is None:
            return hash(self._service_type)

        return (hash(self._service_type) * 397) ^ hash(self._service_key)

    @override
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ServiceIdentifier):
            return NotImplemented

        if self.service_key is None and value.service_key is None:
            return self.service_type == value.service_type

        if self.service_key is not None and value.service_key is not None:
            return (
                self.service_type == value.service_type
                and self.service_key == value.service_key
            )

        return False

    @override
    def __str__(self) -> str:
        if self.service_key is None:
            return str(self.service_type)

        return f"({self.service_key}, {self.service_type})"



from dataclasses import dataclass
from typing import Final, final, override

from wirio._service_lookup._async_concurrent_dictionary import AsyncConcurrentDictionary
from wirio._service_lookup._async_factory_call_site import AsyncFactoryCallSite
from wirio._service_lookup._call_site_visitor import CallSiteVisitor
from wirio._service_lookup._constant_call_site import ConstantCallSite
from wirio._service_lookup._constructor_call_site import ConstructorCallSite
from wirio._service_lookup._service_call_site import ServiceCallSite
from wirio._service_lookup._service_provider_call_site import ServiceProviderCallSite
from wirio._service_lookup._sync_factory_call_site import SyncFactoryCallSite
from wirio._service_lookup._typed_type import TypedType
from wirio._service_lookup.service_cache_key import ServiceCacheKey
from wirio.abstractions.service_scope import ServiceScope
from wirio.abstractions.service_scope_factory import ServiceScopeFactory
from wirio.exceptions import (
    DirectScopedResolvedFromRootError,
    ScopedInSingletonError,
    ScopedResolvedFromRootError,
)


@final
@dataclass
class _CallSiteValidatorState:
    singleton: ServiceCallSite | None


@final
class CallSiteValidator(CallSiteVisitor[_CallSiteValidatorState, TypedType | None]):
    # Keys are services being resolved via `get_service`, values - first scoped service in their call site tree
    _scoped_services: Final[
        AsyncConcurrentDictionary[ServiceCacheKey, TypedType | None]
    ]

    def __init__(self) -> None:
        self._scoped_services = AsyncConcurrentDictionary()

    @override
    async def _visit_call_site(
        self, call_site: ServiceCallSite, argument: _CallSiteValidatorState
    ) -> TypedType | None:
        # First, check if we have encountered this call site before to prevent visiting call site trees that have already been visited
        # If first_scoped_service_in_call_site_tree is null there are no scoped dependencies in this service's call site tree
        # If first_scoped_service_in_call_site_tree has a value, it contains the first scoped service in this service's call site tree

        first_scoped_service_in_call_site_tree = self._scoped_services.get(
            call_site.cache.key
        )

        if first_scoped_service_in_call_site_tree is None:
            # This call site wasn't cached yet, walk the tree
            first_scoped_service_in_call_site_tree = await super()._visit_call_site(
                call_site, argument
            )

            # Cache the result
            await self._scoped_services.upsert(
                call_site.cache.key, first_scoped_service_in_call_site_tree
            )

        # If there is a scoped service in the call site tree, make sure we are not resolving it from a singleton
        if (
            first_scoped_service_in_call_site_tree is not None
            and argument.singleton is not None
        ):
            raise ScopedInSingletonError(
                call_site.service_type, argument.singleton.service_type
            )

        return first_scoped_service_in_call_site_tree

    async def _visit_root_cache(
        self, call_site: ServiceCallSite, argument: _CallSiteValidatorState
    ) -> TypedType | None:
        argument.singleton = call_site
        return await self._visit_call_site_main(call_site, argument)

    async def _visit_scope_cache(
        self, call_site: ServiceCallSite, argument: _CallSiteValidatorState
    ) -> TypedType | None:
        # We are fine with having ServiceScopeFactory requested by singletons
        if call_site.service_type == TypedType.from_type(ServiceScopeFactory):
            return None

        await self._visit_call_site_main(call_site, argument)
        return call_site.service_type

    @override
    async def _visit_constructor(
        self,
        constructor_call_site: ConstructorCallSite,
        argument: _CallSiteValidatorState,
    ) -> TypedType | None:
        result: TypedType | None = None

        for parameter_call_site in constructor_call_site.parameter_call_sites:
            if parameter_call_site is not None:
                scoped = await self._visit_call_site(parameter_call_site, argument)

                if scoped is not None and result is None:
                    result = scoped

        return result

    @override
    def _visit_constant(
        self, constant_call_site: ConstantCallSite, argument: _CallSiteValidatorState
    ) -> TypedType | None:
        return None

    @override
    async def _visit_sync_factory(
        self,
        sync_factory_call_site: SyncFactoryCallSite,
        argument: _CallSiteValidatorState,
    ) -> TypedType | None:
        return None

    @override
    async def _visit_async_factory(
        self,
        async_factory_call_site: AsyncFactoryCallSite,
        argument: _CallSiteValidatorState,
    ) -> TypedType | None:
        return None

    @override
    def _visit_service_provider(
        self,
        service_provider_call_site: ServiceProviderCallSite,
        argument: _CallSiteValidatorState,
    ) -> TypedType | None:
        return None

    async def validate_call_site(self, call_site: "ServiceCallSite") -> None:
        default = _CallSiteValidatorState(singleton=None)
        await self._visit_call_site(call_site, default)

    def validate_resolution(
        self, call_site: ServiceCallSite, scope: ServiceScope, root_scope: ServiceScope
    ) -> None:
        if scope is not root_scope:
            return

        scoped_service = self._scoped_services.get(call_site.cache.key)

        if scoped_service is None:
            return

        service_type = call_site.service_type

        if service_type == scoped_service:
            raise DirectScopedResolvedFromRootError(service_type)

        raise ScopedResolvedFromRootError(service_type, scoped_service)



import typing
from collections.abc import Awaitable, Callable, Generator
from contextlib import AbstractAsyncContextManager, contextmanager
from types import TracebackType
from typing import TYPE_CHECKING, Self, final

from wirio._utils._extra_dependencies import ExtraDependencies
from wirio.abstractions.service_scope import ServiceScope
from wirio.exceptions import ServiceContainerNotBuiltError
from wirio.service_collection import ServiceCollection
from wirio.service_lifetime import ServiceLifetime
from wirio.service_provider import ServiceProvider

if TYPE_CHECKING:
    from fastapi import FastAPI

    from wirio.integrations._fastapi_dependency_injection import (
        FastApiDependencyInjection,
    )
else:
    FastAPI = None
    FastApiDependencyInjection = None


@final
class ServiceContainer(
    ServiceCollection, AbstractAsyncContextManager["ServiceContainer"]
):
    """Collection of resolvable services."""

    _service_provider: ServiceProvider | None

    def __init__(self) -> None:
        super().__init__()
        self._service_provider = None

    @typing.override
    def build_service_provider(
        self, validate_scopes: bool = True, validate_on_build: bool = True
    ) -> ServiceProvider:
        if self._service_provider is not None:
            return self._service_provider

        return super().build_service_provider(validate_scopes, validate_on_build)

    @property
    def service_provider(self) -> ServiceProvider | None:
        return self._service_provider

    async def get[TService](self, service_type: type[TService]) -> TService:
        service_provider = await self._get_service_provider()
        return await service_provider.get_required_service(service_type)

    async def try_get[TService](self, service_type: type[TService]) -> TService | None:
        service_provider = await self._get_service_provider()
        return await service_provider.get_service(service_type)

    async def get_keyed[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService:
        service_provider = await self._get_service_provider()
        return await service_provider.get_required_keyed_service(
            service_key=service_key, service_type=service_type
        )

    async def try_get_keyed[TService](
        self, service_key: object | None, service_type: type[TService]
    ) -> TService | None:
        service_provider = await self._get_service_provider()
        return await service_provider.get_keyed_service(
            service_key=service_key, service_type=service_type
        )

    def create_scope(self) -> ServiceScope:
        """Create a new :class:`ServiceScope` that can be used to resolve scoped services."""
        if self._service_provider is None:
            self._service_provider = self.build_service_provider()

        return self._service_provider.create_scope()

    async def aclose(self) -> None:
        if self._service_provider is not None:
            await self._service_provider.__aexit__(None, None, None)

        self._service_provider = None

    @contextmanager
    def override(
        self, service_type: type, implementation_instance: object | None
    ) -> Generator[None]:
        """Override a service registration within the context manager scope.

        It can be used to temporarily replace a service for testing specific scenarios. Don't use it in production.
        """
        self._ensure_service_container_is_built()
        assert self._service_provider is not None

        with self._service_provider.override_service(
            service_type=service_type,
            implementation_instance=implementation_instance,
        ):
            yield

    @contextmanager
    def override_keyed(
        self,
        service_key: object | None,
        service_type: type,
        implementation_instance: object | None,
    ) -> Generator[None]:
        """Override a keyed service registration within the context manager scope.

        It can be used to temporarily replace a service for testing specific scenarios. Don't use it in production.
        """
        self._ensure_service_container_is_built()
        assert self._service_provider is not None

        with self._service_provider.override_keyed_service(
            service_key=service_key,
            service_type=service_type,
            implementation_instance=implementation_instance,
        ):
            yield

    @typing.override
    def configure_fastapi(self, app: FastAPI) -> None:
        """Configure the FastAPI application to use dependency injection using the services from this service container."""
        self._import_fastapi()
        FastApiDependencyInjection.setup(app, self)

    @typing.override
    def _import_fastapi(self) -> None:
        ExtraDependencies.import_fastapi()
        global FastApiDependencyInjection  # noqa: PLW0603
        from wirio.integrations._fastapi_dependency_injection import (  # noqa: PLC0415
            FastApiDependencyInjection,
        )

    @typing.override
    def _add[TService](
        self,
        lifetime: ServiceLifetime,
        service_type: type[TService] | None,
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None,
        implementation_type: type | None,
        implementation_instance: object | None,
        service_key: object | None,
        auto_activate: bool,
    ) -> None:
        super()._add(
            lifetime=lifetime,
            service_type=service_type,
            implementation_factory=implementation_factory,
            implementation_type=implementation_type,
            implementation_instance=implementation_instance,
            service_key=service_key,
            auto_activate=auto_activate,
        )

        if self._service_provider is not None:
            added_descriptor = self._descriptors[-1]
            self._service_provider.add_descriptor(added_descriptor)

    async def _get_service_provider(self) -> ServiceProvider:
        if self._service_provider is None:
            self._service_provider = self.build_service_provider()
            await self._service_provider.__aenter__()

        return self._service_provider

    def _ensure_service_container_is_built(self) -> None:
        if self._service_provider is None:
            raise ServiceContainerNotBuiltError

    @typing.override
    async def __aenter__(self) -> Self:
        service_provider = await self._get_service_provider()
        await service_provider.fully_initialize_if_not_fully_initialized()
        return self

    @typing.override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        assert self._service_provider is not None
        await self._service_provider.__aexit__(exc_type, exc_val, exc_tb)
        self._service_provider = None



import inspect
import typing
from collections.abc import Awaitable, Callable, Iterator
from typing import TYPE_CHECKING, Final, cast, overload

from wirio._service_lookup._typed_type import TypedType
from wirio._utils._extra_dependencies import ExtraDependencies
from wirio.exceptions import (
    NoKeyedSingletonServiceRegisteredError,
    NoSingletonServiceRegisteredError,
)
from wirio.service_descriptor import ServiceDescriptor
from wirio.service_lifetime import ServiceLifetime
from wirio.service_provider import ServiceProvider
from wirio.wirio_undefined import WirioUndefined

if TYPE_CHECKING:
    from fastapi import FastAPI

    from wirio.integrations._fastapi_dependency_injection import (
        FastApiDependencyInjection,
    )
else:
    FastAPI = None
    FastApiDependencyInjection = None


class ServiceCollection:
    """Collection of service descriptors provided during configuration."""

    _descriptors: Final[list[ServiceDescriptor]]

    def __init__(self) -> None:
        self._descriptors = []
        self._validate_on_build = True

    def build_service_provider(
        self, validate_scopes: bool = True, validate_on_build: bool = True
    ) -> ServiceProvider:
        """Create a :class:`ServiceProvider` containing services from the provided :class:`ServiceCollection`."""
        return ServiceProvider(
            descriptors=self._descriptors,
            validate_scopes=validate_scopes,
            validate_on_build=validate_on_build,
        )

    @overload
    def add_transient[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_transient[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    def add_transient[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory_or_implementation_type_or_none: Callable[
            ..., Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | None = None,
        /,
    ) -> None:
        """Add a transient service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.TRANSIENT,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_none,
        )

    @overload
    def add_singleton[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_instance: object,
        /,
    ) -> None: ...

    def add_singleton[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory_or_implementation_type_or_implementation_instance_or_none: Callable[
            ..., Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | object
        | None = None,
        /,
    ) -> None:
        """Add a singleton service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SINGLETON,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_implementation_instance_or_none,
        )

    @overload
    def add_scoped[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    def add_scoped[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory_or_implementation_type_or_none: Callable[
            ..., Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | None = None,
        /,
    ) -> None:
        """Add a scoped service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SCOPED,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_none,
        )

    @overload
    def add_keyed_transient[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_transient[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_transient[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_transient[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_transient[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_transient[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    def add_keyed_transient[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type_or_implementation_factory: type[TService]
        | Callable[[TKey | None, *TFactoryParameter], Awaitable[TService]]
        | Callable[[TKey | None, *TFactoryParameter], TService],
        implementation_factory_or_implementation_type_or_none: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ]
        | Callable[[TKey | None, *TFactoryParameter], TService]
        | type
        | None = None,
        /,
    ) -> None:
        """Add a keyed transient service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.TRANSIENT,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_none,
            service_key=service_key,
        )

    @overload
    def add_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    @overload
    def add_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_instance: object,
        /,
    ) -> None: ...

    def add_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type_or_implementation_factory: type[TService]
        | Callable[[TKey | None, *TFactoryParameter], Awaitable[TService]]
        | Callable[[TKey | None, *TFactoryParameter], TService],
        implementation_factory_or_implementation_type_or_implementation_instance_or_none: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ]
        | Callable[[TKey | None, *TFactoryParameter], TService]
        | type
        | object
        | None = None,
        /,
    ) -> None:
        """Add a keyed singleton service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SINGLETON,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_implementation_instance_or_none,
            service_key=service_key,
        )

    @overload
    def add_keyed_scoped[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_scoped[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_scoped[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_scoped[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_keyed_scoped[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_keyed_scoped[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    def add_keyed_scoped[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type_or_implementation_factory: type[TService]
        | Callable[[TKey | None, *TFactoryParameter], Awaitable[TService]]
        | Callable[[TKey | None, *TFactoryParameter], TService],
        implementation_factory_or_implementation_type_or_none: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | None = None,
        /,
    ) -> None:
        """Add a keyed scoped service."""
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SCOPED,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_none,
            service_key=service_key,
        )

    @overload
    def add_auto_activated_singleton[TService](
        self, service_type: type[TService], /
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_singleton[TService](
        self,
        service_type: type[TService],
        implementation_instance: object,
        /,
    ) -> None: ...

    def add_auto_activated_singleton[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory_or_implementation_type_or_implementation_instance_or_none: Callable[
            ..., Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | object
        | None = None,
        /,
    ) -> None:
        """Add an auto-activated singleton service.

        An auto-activated singleton service is instantiated when the service provider is built (eagerly), rather than when it's first requested (lazily).
        """
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SINGLETON,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_implementation_instance_or_none,
            auto_activate=True,
        )

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type: type[TService],
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        implementation_factory: Callable[[TKey | None, *TFactoryParameter], TService],
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_type: type,
        /,
    ) -> None: ...

    @overload
    def add_auto_activated_keyed_singleton[TKey, TService](
        self,
        service_key: TKey | None,  # pyright: ignore[reportInvalidTypeVarUse]
        service_type: type[TService],
        implementation_instance: object,
        /,
    ) -> None: ...

    def add_auto_activated_keyed_singleton[TKey, TService, *TFactoryParameter](
        self,
        service_key: TKey | None,
        service_type_or_implementation_factory: type[TService]
        | Callable[[TKey | None, *TFactoryParameter], Awaitable[TService]]
        | Callable[[TKey | None, *TFactoryParameter], TService],
        implementation_factory_or_implementation_type_or_implementation_instance_or_none: Callable[
            [TKey | None, *TFactoryParameter], Awaitable[TService]
        ]
        | Callable[[TKey | None, *TFactoryParameter], TService]
        | type
        | object
        | None = None,
        /,
    ) -> None:
        """Add an auto-activated keyed singleton service.

        An auto-activated keyed singleton service is instantiated when the service provider is built (eagerly), rather than when it's first requested (lazily).
        """
        self._add_from_overloaded_constructor(
            lifetime=ServiceLifetime.SINGLETON,
            service_type_or_implementation_factory=service_type_or_implementation_factory,
            implementation_factory_or_implementation_type_or_implementation_instance_or_none=implementation_factory_or_implementation_type_or_implementation_instance_or_none,
            service_key=service_key,
            auto_activate=True,
        )

    def enable_singleton_auto_activation(self, service_type: type) -> None:
        """Mark a registered singleton service as auto-activated.

        An auto-activated singleton service is instantiated when the service provider is built (eagerly), rather than when it's first requested (lazily).
        """
        typed_service_type = TypedType.from_type(service_type)
        is_descriptor_found = False

        for descriptor in self._descriptors:
            if (
                descriptor.service_type == typed_service_type
                and descriptor.lifetime == ServiceLifetime.SINGLETON
            ):
                is_descriptor_found = True
                descriptor.auto_activate = True

        if not is_descriptor_found:
            raise NoSingletonServiceRegisteredError(typed_service_type)

    def enable_keyed_singleton_auto_activation(
        self, service_key: object | None, service_type: type
    ) -> None:
        """Mark a registered keyed singleton service as auto-activated.

        An auto-activated keyed singleton service is instantiated when the service provider is built (eagerly), rather than when it's first requested (lazily).
        """
        typed_service_type = TypedType.from_type(service_type)
        is_descriptor_found = False

        for descriptor in self._descriptors:
            if (
                descriptor.is_keyed_service
                and descriptor.lifetime == ServiceLifetime.SINGLETON
                and descriptor.service_type == typed_service_type
                and descriptor.service_key == service_key
            ):
                is_descriptor_found = True
                descriptor.auto_activate = True

        if not is_descriptor_found:
            raise NoKeyedSingletonServiceRegisteredError(
                service_type=typed_service_type, service_key_type=type(service_key)
            )

    def configure_fastapi(self, app: FastAPI) -> None:
        """Configure the FastAPI application to use dependency injection using the services from this service collection."""
        self._import_fastapi()
        FastApiDependencyInjection.setup(app, self)

    def _add_from_overloaded_constructor[TService](
        self,
        lifetime: ServiceLifetime,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory_or_implementation_type_or_implementation_instance_or_none: Callable[
            ..., Awaitable[TService]
        ]
        | Callable[..., TService]
        | type
        | object
        | None = None,
        service_key: object | None = WirioUndefined.INSTANCE,
        auto_activate: bool = False,
    ) -> None:
        service_type_to_add: type[TService] | None = None
        implementation_factory_to_add: (
            Callable[..., Awaitable[TService]] | Callable[..., TService] | None
        ) = None
        implementation_type_to_add: type | None = None
        implementation_instance_to_add: object | None = None

        if isinstance(service_type_or_implementation_factory, type):
            service_type_to_add = service_type_or_implementation_factory

        if isinstance(
            implementation_factory_or_implementation_type_or_implementation_instance_or_none,
            type,
        ):
            implementation_type_to_add = implementation_factory_or_implementation_type_or_implementation_instance_or_none
        elif (
            service_type_to_add is not None
            and implementation_factory_or_implementation_type_or_implementation_instance_or_none
            is not None
        ):
            if callable(
                implementation_factory_or_implementation_type_or_implementation_instance_or_none
            ):
                implementation_factory_to_add = cast(
                    "Callable[..., Awaitable[TService]] | Callable[..., TService]",
                    implementation_factory_or_implementation_type_or_implementation_instance_or_none,
                )
            else:
                implementation_instance_to_add = implementation_factory_or_implementation_type_or_implementation_instance_or_none
        elif (
            service_type_to_add is None
            and implementation_factory_or_implementation_type_or_implementation_instance_or_none
            is None
        ):
            implementation_factory_to_add = service_type_or_implementation_factory

        self._add(
            lifetime=lifetime,
            service_type=service_type_to_add,
            implementation_factory=implementation_factory_to_add,
            implementation_type=implementation_type_to_add,
            implementation_instance=implementation_instance_to_add,
            service_key=service_key,
            auto_activate=auto_activate,
        )

    def _add[TService](  # noqa: PLR0913
        self,
        lifetime: ServiceLifetime,
        service_type: type[TService] | None,
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None,
        implementation_type: type | None,
        implementation_instance: object | None,
        service_key: object | None,
        auto_activate: bool,
    ) -> None:
        provided_service_type = self._get_provided_service_type(
            service_type, implementation_factory
        )
        is_service_key_provided = service_key is not WirioUndefined.INSTANCE
        service_key_to_add = service_key if is_service_key_provided else None
        service_descriptor: ServiceDescriptor | None = None

        if implementation_instance is not None:
            service_descriptor = ServiceDescriptor.from_implementation_instance(
                service_type=provided_service_type,
                implementation_instance=implementation_instance,
                service_key=service_key_to_add,
                lifetime=lifetime,
                auto_activate=auto_activate,
            )
        elif implementation_factory is None:
            if implementation_type is not None and (
                not issubclass(implementation_type, provided_service_type)
                or provided_service_type == implementation_type
            ):
                error_message = (
                    f"{implementation_type} is not subclass of {provided_service_type}"
                )
                raise TypeError(error_message)

            implementation_type_to_add = (
                implementation_type
                if implementation_type is not None
                else provided_service_type
            )
            service_descriptor = ServiceDescriptor.from_implementation_type(
                service_type=provided_service_type,
                implementation_type=implementation_type_to_add,
                service_key=service_key_to_add,
                lifetime=lifetime,
                auto_activate=auto_activate,
            )
        elif inspect.iscoroutinefunction(implementation_factory):
            if is_service_key_provided:
                service_descriptor = (
                    ServiceDescriptor.from_keyed_async_implementation_factory(
                        service_type=provided_service_type,
                        implementation_factory=implementation_factory,
                        service_key=service_key_to_add,
                        lifetime=lifetime,
                        auto_activate=auto_activate,
                    )
                )
            else:
                service_descriptor = (
                    ServiceDescriptor.from_async_implementation_factory(
                        service_type=provided_service_type,
                        implementation_factory=implementation_factory,
                        lifetime=lifetime,
                        auto_activate=auto_activate,
                    )
                )
        else:  # noqa: PLR5501
            if is_service_key_provided:
                service_descriptor = (
                    ServiceDescriptor.from_keyed_sync_implementation_factory(
                        service_type=provided_service_type,
                        implementation_factory=implementation_factory,
                        service_key=service_key_to_add,
                        lifetime=lifetime,
                        auto_activate=auto_activate,
                    )
                )
            else:
                service_descriptor = ServiceDescriptor.from_sync_implementation_factory(
                    service_type=provided_service_type,
                    implementation_factory=implementation_factory,
                    lifetime=lifetime,
                    auto_activate=auto_activate,
                )

        assert service_descriptor is not None
        self._descriptors.append(service_descriptor)

    def _get_provided_service_type[TService](
        self,
        service_type: type[TService] | None = None,
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
    ) -> type:
        if service_type is not None:
            return service_type

        assert implementation_factory is not None

        type_hints: dict[str, type] = typing.get_type_hints(implementation_factory)
        return_type = type_hints.get("return")

        if return_type is None:
            error_message = "Missing return type hints from 'implementation_factory'"
            raise ValueError(error_message)

        return return_type

    def _import_fastapi(self) -> None:
        ExtraDependencies.import_fastapi()
        global FastApiDependencyInjection  # noqa: PLW0603
        from wirio.integrations._fastapi_dependency_injection import (  # noqa: PLC0415
            FastApiDependencyInjection,
        )

    def __iter__(self) -> Iterator[ServiceDescriptor]:
        return iter(self._descriptors)



from collections.abc import Awaitable, Callable
from typing import ClassVar, final, override

from wirio._service_lookup._call_site_runtime_resolver import (
    CallSiteRuntimeResolver,
)
from wirio._service_lookup._service_call_site import (
    ServiceCallSite,
)
from wirio._service_lookup._service_provider_engine import (
    ServiceProviderEngine,
)
from wirio.service_provider_engine_scope import (
    ServiceProviderEngineScope,
)


@final
class RuntimeServiceProviderEngine(ServiceProviderEngine):
    INSTANCE: ClassVar["RuntimeServiceProviderEngine"]

    @override
    def realize_service(
        self, call_site: ServiceCallSite
    ) -> Callable[[ServiceProviderEngineScope], Awaitable[object | None]]:
        def _create_realize_service(
            scope: ServiceProviderEngineScope,
        ) -> Awaitable[object | None]:
            return CallSiteRuntimeResolver.INSTANCE.resolve(call_site, scope)

        return _create_realize_service


RuntimeServiceProviderEngine.INSTANCE = RuntimeServiceProviderEngine()



class ExtraDependencies:
    @staticmethod
    def is_fastapi_installed() -> bool:
        try:
            import fastapi  # pyright: ignore[reportUnusedImport] # noqa: F401, PLC0415
        except ImportError:
            return False

        return True

    @staticmethod
    def import_fastapi() -> None:
        try:
            from fastapi import (  # noqa: PLC0415
                FastAPI,  # pyright: ignore[reportUnusedImport]  # noqa: F401
            )
        except ImportError as error:
            error_message = "FastAPI is not installed. Please, run 'uv add wirio[fastapi]' to install the required dependencies."
            raise ImportError(error_message) from error
